#!/usr/bin/python3

import sys
import subprocess
import os, stat
import re
import logging
import json
import datetime
import yaml


logging.basicConfig(format='%(asctime)s %(message)s', level=logging.INFO)
raise_error = 0

# Location for dumps.
db_dir = "/root/docker-db-dump"

# Optionally hang on to older dumps, but we want to keep almost a days worth, even if we're 
# dumping backups hourly or similar.
keep_seconds = 85000

compression_command = "pigz"
compression_extension = "gz"
decompression_command = "zcat"

config_file = "/etc/mythic/docker-db.yaml"

config = None
if os.path.exists(config_file):
    with open(config_file) as f:
        config = yaml.safe_load(f)

def detect_db(container_id):
    processes = subprocess.check_output(['docker', 'top', container_id, '-eo', 'pid,cmd']).decode('utf-8')
    my_re = re.search("(mysqld(?!ump))", processes)
    pg_re = re.search("(postgres:)", processes)
    if pg_re and my_re:
        return "error"
    if pg_re:
        return "postgres"
    if my_re:
        return "mysql"
    return "none"


def find_config(container_id):
    container_json = subprocess.check_output(['docker', 'inspect', container_id]).decode('utf-8')
    j = json.loads(container_json)
    container_config = j[0]['Config']
    container_name = j[0]['Name'].lstrip('/')

    if config is not None and 'containers' in config:
        if container_name in config['containers']:
            logging.debug("Returning config for: %s", container_name)
            return [ config['containers'][container_name]['username'],config['containers'][container_name]['password'], container_name ]

    container_env = container_config['Env']
    logging.debug(container_env)
    db_user = None
    db_pass = None
    root_pass = None
    for e in container_env:
        logging.debug(e)
        res = re.search("_([A-Z]+)=(.*)$", e)
        if res:
            logging.debug("Got %s %s",res.group(1),res.group(2))
            if res.group(1) == "USER":
                db_user = res.group(2)
            if res.group(1) == "USERNAME":
                db_user = res.group(2)
            if res.group(1) == "PASS":
                db_pass = res.group(2)
            if res.group(1) == "PASSWORD":
                db_pass = res.group(2)

        # Some containers have MYSQL_PASSWORD for an unprivileged user, and MYSQL_ROOT_PASSWORD
        res = re.search("_ROOT_PASSWORD=(.*)$", e)
        if res:
            logging.debug("Overriding root password %s",res.group(1))
            root_pass = res.group(1)

    if root_pass is not None:
        db_pass=root_pass
        db_user="root"

    logging.debug("Guessing username: %s", db_user)
    logging.debug("Guessing password: %s", db_pass)
    return [db_user, db_pass, container_name]


def backup_db(container_id, runcmd, filename):
    global raise_error
    logging.debug("Running: %s", " ".join(runcmd))

    f = open(filename, 'wb')

    dump = subprocess.Popen(runcmd, stdout=subprocess.PIPE)
    z = subprocess.Popen(compression_command, stdin=dump.stdout,stdout=f)
    z.communicate()

    logging.debug("Return code: %d", z.returncode)
    if z.returncode != 0:
        logging.warning("Non-zero return code! %d", z.returncode)
        raise_error = 1

    dump.wait()
    logging.debug("Return code: %d", dump.returncode)
    if dump.returncode != 0:
        logging.warning("Non-zero return code! %d", dump.returncode)
        raise_error = 1

    f.close()


def get_backupname(container_id, database):
    # mysql.CONTAINER.2000-02-29T12:34:56.Tue.sql.(gz|lzo|lz4|zst)
    now = datetime.datetime.now()
    format = "%Y-%m-%dT%H:%M:%S.%a"
    tstamp = now.strftime(format)
    filename = "%s/%s.%s.%s.sql.%s" % (db_dir, database, container_id, tstamp, compression_extension)
    logging.info("Filename: %s", filename)
    return filename


def backup_pg(container_id):
    logging.debug("Starting postgres backups for %s", container_id)
    creds = find_config(container_id)
    pg_user = "postgres" if creds[0] is None else creds[0]

    runcmd = ['docker', 'exec', container_id, 'pg_dumpall', '-U', pg_user]

    backup_db(container_id, runcmd, get_backupname(container_id, "postgres"))
    logging.debug("Finished postgres backups for %s", container_id)


def check_mysql(filename):
    cmd = " ".join([decompression_command, filename, '|', 'tail', '-1'])
    mycheck = subprocess.check_output(cmd,shell=True).decode('utf-8')
    logging.debug("mysql result: %s", mycheck)

    if re.match("-- Dump completed",mycheck):
        return True

    logging.warning("mysql check result: %s", mycheck)
    return False

def backup_mysql(container_id):
    global raise_error
    logging.debug("Starting mysql backups for %s", container_id)
    creds = find_config(container_id)
    my_user = "root" if creds[0] is None else creds[0]
    my_pass = creds[1]

    runcmd = ['docker', 'exec', container_id, 'mysqldump', '--all-databases', '--single-transaction','--events','--flush-logs', '-u', my_user]

    container_name = creds[2]
    if config is not None and 'containers' in config:
        if container_name in config['containers'] and 'mysql_args' in config['containers'][container_name]:
            logging.debug("Custom mysql config for: %s", container_name)
            runcmd = ['docker', 'exec', container_id, 'mysqldump', '-u', my_user] + config['containers'][container_name]['mysql_args']

    if my_pass is not None:
        runcmd += ["--password=" + my_pass]

    filename = get_backupname(container_id, "mysql")
    backup_db(container_id, runcmd, filename)
    mycheck = check_mysql(filename)
    if not mycheck:
        raise_error = 1
        logging.warning("mysql dump truncated, missing 'Dump completed'")
    logging.debug("Finished mysql backups for %s", container_id)


# Prune older backups
def clean_backups():
    logging.debug("Tidying backups in %s", db_dir)
    dirs = os.listdir(db_dir) 
    currenttime = datetime.datetime.now().timestamp()
    for file in dirs:
        logging.debug("Found file %s", file)
        if re.match("(postgres|mysql)\.[^\.]+\.[0-9-]{10}T[0-9:]{8}\.\S+\.sql.(gz|lzo|lz4|zst)$", file):
            fullpath = os.path.join(db_dir, file)
            age = currenttime - os.path.getmtime(fullpath)
            logging.debug("regex match, age: %d", age)
            if age > keep_seconds:
                logging.debug("Removing old backup, %s", file)
                os.unlink(fullpath)


# Create dump dir and make sure it's definitely locked down...
if not os.path.exists(db_dir): 
    os.mkdir(db_dir, stat.S_IRWXU)
os.chmod(db_dir, stat.S_IRWXU)

# Check each container for a running DB, back it up.
containers = subprocess.check_output(['docker', 'ps', '-q']).decode('utf-8')
for container in containers.split('\n'):
    if container:
        logging.info("Checking container: %s", container)
        db = detect_db(container)
        if db == "mysql":
            logging.info("Found mysql in %s", container)
            backup_mysql(container)
        elif db == "postgres":
            logging.info("Found postgres in %s", container)
            backup_pg(container)
        elif db == "error":
            logging.warning("Error detecting database in %s", container)
            raise_error = 1
        else:
            logging.info("no DB found in %s", container)


# We want to delay an error exit so we attempt all databases before 
# complaining - in this case a non-zero exit code for mythic-backup.
if raise_error > 0:
    logging.warning("raise_error set, something went wrong!")
    sys.exit(1)
else:
    logging.info("Clean exit \o/")
    # Don't remove older files if we're erroring. 
    clean_backups()
    sys.exit(0)
